@@ -14,12 +14,15 @@
 #define FAN_CONTROL_FILE "/proc/acpi/ibm/fan"
 #define TEMP_INVALID INT_MIN
 #define TEMP_MIN INT_MIN + 1
+#define NS_IN_SEC 1000000000L  // 1 second in nanoseconds
+#define THRESHOLD_NS 200000000 // 0.2 seconds
 
 #define STR_HELPER(x) #x
 #define STR(x) STR_HELPER(x)
 #define DEFAULT_WATCHDOG_SECS 120
 #define S_DEFAULT_WATCHDOG_SECS STR(DEFAULT_WATCHDOG_SECS)
 
+#define info(fmt, ...) fprintf(stderr, "[INF] " fmt, ##__VA_ARGS__)
 #define err(fmt, ...) fprintf(stderr, "[ERR] " fmt, ##__VA_ARGS__)
 #define max(x, y) ((x) > (y) ? (x) : (y))
 #define expect(x)                                                              \
@@ -31,10 +34,13 @@
         }                                                                      \
     } while (0)
 
+#define CONFIG_MAX_STRLEN 15
+#define S_CONFIG_MAX_STRLEN STR(CONFIG_MAX_STRLEN)
+
 /* Must be highest to lowest temp */
 enum FanLevel { FAN_MAX, FAN_MED, FAN_LOW, FAN_OFF, FAN_INVALID };
 struct Rule {
-    const char *tpacpi_level;
+    char tpacpi_level[CONFIG_MAX_STRLEN + 1];
     int threshold;
     const char *name;
 };
@@ -47,7 +53,7 @@ static struct Rule rules[] = {
 
 static struct timespec last_watchdog_ping = {0, 0};
 static time_t watchdog_secs = DEFAULT_WATCHDOG_SECS;
-static const unsigned int fan_hysteresis = 10;
+static int temp_hysteresis = 10;
 static const unsigned int tick_hysteresis = 3;
 static char output_buf[512];
 static const struct Rule *current_rule = NULL;
@@ -55,6 +61,11 @@ static volatile sig_atomic_t run = 1;
 static int first_tick = 1; /* Stop running if errors are immediate */
 static glob_t temp_files;
 
+enum resume_state {
+    RESUME_NOT_DETECTED,
+    RESUME_DETECTED,
+};
+
 static void exit_if_first_tick(void) {
     if (first_tick) {
         err("Quitting due to failure during first run\n");
@@ -62,6 +73,36 @@ static void exit_if_first_tick(void) {
     }
 }
 
+static int64_t timespec_diff_ns(const struct timespec *start,
+                                const struct timespec *end) {
+    return ((int64_t)end->tv_sec - (int64_t)start->tv_sec) * NS_IN_SEC +
+           (end->tv_nsec - start->tv_nsec);
+}
+
+static enum resume_state detect_suspend(void) {
+    static struct timespec monotonic_prev, boottime_prev;
+    struct timespec monotonic_now, boottime_now;
+
+    expect(clock_gettime(CLOCK_MONOTONIC, &monotonic_now) == 0);
+    expect(clock_gettime(CLOCK_BOOTTIME, &boottime_now) == 0);
+
+    if (monotonic_prev.tv_sec == 0 && monotonic_prev.tv_nsec == 0) {
+        monotonic_prev = monotonic_now;
+        boottime_prev = boottime_now;
+        return RESUME_NOT_DETECTED;
+    }
+
+    int64_t delta_monotonic = timespec_diff_ns(&monotonic_prev, &monotonic_now);
+    int64_t delta_boottime = timespec_diff_ns(&boottime_prev, &boottime_now);
+
+    monotonic_prev = monotonic_now;
+    boottime_prev = boottime_now;
+
+    return delta_boottime > delta_monotonic + THRESHOLD_NS
+               ? RESUME_DETECTED
+               : RESUME_NOT_DETECTED;
+}
+
 static int glob_err_handler(const char *epath, int eerrno) {
     err("glob: %s: %s\n", epath, strerror(eerrno));
     return 0;
@@ -178,7 +219,7 @@ static int set_fan_level(void) {
             if (tick_penalty) {
                 return 0; /* Must wait longer until able to move down levels */
             }
-            temp_penalty = fan_hysteresis;
+            temp_penalty = temp_hysteresis;
         }
 
         if (rule->threshold < temp_penalty ||
@@ -203,15 +244,24 @@ static int set_fan_level(void) {
 static void maybe_ping_watchdog(void) {
     struct timespec now;
 
+    expect(current_rule);
     expect(clock_gettime(CLOCK_MONOTONIC, &now) == 0);
 
+    if (detect_suspend() == RESUME_DETECTED) {
+        // On resume, some models need a manual fan write again, or they will
+        // revert to "auto".
+        info("Clock jump detected, possible resume. Rewriting fan level\n");
+        write_fan_level(current_rule->tpacpi_level);
+    }
+
     if (now.tv_sec - last_watchdog_ping.tv_sec <
         (watchdog_secs - WATCHDOG_GRACE_PERIOD_SECS)) {
         return;
     }
 
-    expect(current_rule); /* Already set up on first run by set_fan_level */
-    write_fan_level(current_rule->tpacpi_level);
+    // Transitioning from level 0 -> level 0 can cause a brief fan spinup on
+    // some models, so don't reset the timer by write_fan_level().
+    write_watchdog_timeout(watchdog_secs);
 }
 
 #define CONFIG_PATH "/etc/zcfan.conf"
@@ -225,6 +275,17 @@ static void maybe_ping_watchdog(void) {
         }                                                                      \
     } while (0)
 
+#define fscanf_str_for_key(f, pos, name, dest)                                 \
+    do {                                                                       \
+        char val[CONFIG_MAX_STRLEN + 1];                                       \
+        if (fscanf(f, name " %" S_CONFIG_MAX_STRLEN "s ", val) == 1) {         \
+            strncpy(dest, val, CONFIG_MAX_STRLEN);                             \
+            dest[CONFIG_MAX_STRLEN] = '\0';                                    \
+        } else {                                                               \
+            expect(fseek(f, pos, SEEK_SET) == 0);                              \
+        }                                                                      \
+    } while (0)
+
 static void get_config(void) {
     FILE *f;
 
@@ -245,6 +306,10 @@ static void get_config(void) {
         fscanf_int_for_key(f, pos, "med_temp", rules[FAN_MED].threshold);
         fscanf_int_for_key(f, pos, "low_temp", rules[FAN_LOW].threshold);
         fscanf_int_for_key(f, pos, "watchdog_secs", watchdog_secs);
+        fscanf_int_for_key(f, pos, "temp_hysteresis", temp_hysteresis);
+        fscanf_str_for_key(f, pos, "max_level", rules[FAN_MAX].tpacpi_level);
+        fscanf_str_for_key(f, pos, "med_level", rules[FAN_MED].tpacpi_level);
+        fscanf_str_for_key(f, pos, "low_level", rules[FAN_LOW].tpacpi_level);
         if (ftell(f) == pos) {
             while ((ch = fgetc(f)) != EOF && ch != '\n') {}
         }
@@ -296,7 +361,8 @@ int main(int argc, char *argv[]) {
 
     if (!full_speed_supported()) {
         err("level \"full-speed\" not supported, using level 7\n");
-        rules[FAN_MAX].tpacpi_level = "7";
+        strncpy(rules[FAN_MAX].tpacpi_level, "7", CONFIG_MAX_STRLEN);
+        rules[FAN_MAX].tpacpi_level[CONFIG_MAX_STRLEN] = '\0';
     }
 
     write_watchdog_timeout(watchdog_secs);
